(define-module (ordo interceptor)
  #:use-module (ice-9 exceptions)
  #:use-module (logging logger)
  #:use-module (srfi srfi-1)  ; list utils
  #:use-module (srfi srfi-9)  ; records
  #:use-module (srfi srfi-26) ; cut
  #:use-module (srfi srfi-69) ; hash tables
  #:use-module (srfi srfi-71) ; extended let
  #:export (interceptor
            init-context
            context-connection
            set-context-connection!
            context-error
            set-context-error!
            context-suppressed
            context-vars
            set-context-vars!
            var-ref
            var-set!
            var-delete!
            terminate-when
            execute))

(define (check-var-name name)
  (unless (keyword? name)
    (raise-exception (make-exception
                      (make-assertion-failure)
                      (make-exception-with-message "Variable name should be a keyword")
                      (make-exception-with-irritants name)))))

(define-record-type <context>
  (make-context connection vars stack queue terminators error suppressed)
  context?
  (connection context-connection set-context-connection!)
  (vars context-vars set-context-vars!)
  (stack context-stack set-context-stack!)
  (queue context-queue set-context-queue!)
  (terminators context-terminators set-context-terminators!)
  (error context-error set-context-error!)
  (suppressed context-suppressed set-context-suppressed!))

(define* (init-context #:key conn (vars '()))
  "Initialize a context with optional connection and vars."
  (for-each check-var-name (map car vars))
  (make-context
   ;; connection
   conn
   ;; vars
   (alist->hash-table vars equal?)
   ;; stack
   '()
   ;; queue
   '()
   ;; terminators
   '()
   ;; error
   #f
   ;; suppressed errors
   '()))

(define (var-set! ctx name value)
  (check-var-name name)
  (log-msg 'DEBUG "Setting variable " name " to " value)
  (hash-table-set! (context-vars ctx) name value))

(define* (var-ref ctx name #:optional default)
  (check-var-name name)
  (log-msg 'DEBUG "Getting variable " name " with default " default)
  (hash-table-ref/default (context-vars ctx) name default))

(define (var-delete! ctx name)
  (check-var-name name)
  (log-msg 'DEBUG "Deleting variable " name)
  (hash-table-delete! (context-vars ctx) name))

(define-record-type <interceptor>
  (make-interceptor name enter leave error register)
  interceptor?
  (name interceptor-name)
  (enter interceptor-enter)
  (leave interceptor-leave)
  (error interceptor-error)
  (register interceptor-register))

(define* (interceptor name #:key enter leave error register)
  "Create an interceptor with optional enter, leave, and error functions."
  (make-interceptor name enter leave error register))

(define-exception-type &interceptor-error &error
  make-interceptor-error
  interceptor-error?
  (interceptor-name interceptor-error-interceptor-name)
  (stage interceptor-error-stage)
  (cause interceptor-error-cause))

(define (enqueue ctx interceptors)
  "Add interceptors to the context."
  (unless (every interceptor? interceptors)
    (error "invalid interceptors"))
  (set-context-queue! ctx interceptors))

(define (terminate ctx)
  "Remove all remaining interceptors from the queue, short-circuiting the
   enter stage and running the leave stage."
  (set-context-queue! ctx '()))

(define (check-terminators ctx)
  "Check the context terminators and possibly trigger early termination."
  (let loop ((terminators (context-terminators ctx)))
    (unless (null? terminators)
      (let ((t (car terminators)))
        (if (t ctx)
            (terminate ctx)
            (loop (cdr terminators)))))))

(define (try-enter ctx t)
  "Run the interceptor's #:enter function."
  (let ((handler (interceptor-enter t)))
    (when handler
      (log-msg 'NOTICE "Running #:enter function for " (interceptor-name t))
      (with-exception-handler
          (lambda (e)
            (set-context-error! ctx
                                (make-interceptor-error (interceptor-name t) #:enter e)))
        (lambda ()
          (let ((result (handler ctx)))
            (and=> (interceptor-register t) (cut var-set! ctx <> result))))
        #:unwind? #t))))

(define (try-leave ctx t)
  "Run the interceptor's #:leave function."
  (let ((handler (interceptor-leave t)))
    (when handler
      (log-msg 'NOTICE "Running #:leave function for " (interceptor-name t))
      (with-exception-handler
          (lambda (e)
            (set-context-error! ctx
                                (make-interceptor-error (interceptor-name t) #:leave e)))
        (lambda () (handler ctx))
        #:unwind? #t))))

(define (try-error ctx t err)
  "Run the interceptor's #:error function."
  (let ((handler (interceptor-error t)))
    (when handler
      (log-msg 'NOTICE "Running #:error function for " (interceptor-name t))
      (with-exception-handler
          (lambda (e)
            (log-msg 'WARN "error handler for interceptor '" (interceptor-name t) "' threw error: " e)
            (set-context-suppressed! ctx
                                     (cons (make-interceptor-error (interceptor-name t) #:error e)
                                           (context-suppressed ctx))))
        (lambda () (handler ctx))
        #:unwind? #t))))

(define (execute-leave ctx)
  "Run all the #:leave functions in the queue."
  (unless (null? (context-queue ctx))
    (let ((t (car (context-queue ctx)))
          (err (context-error ctx)))
      ;; Run the error or leave handler, according to whether or not we are
      ;; handling an error
      (if err
          (try-error ctx t err)
          (try-leave ctx t))
      ;; Remove the current interceptor from the queue and add it to the stack
      (set-context-stack! ctx (cons t (context-stack ctx)))
      (set-context-queue! ctx (cdr (context-queue ctx)))
      ;; Carry on down the chain
      (execute-leave ctx))))

(define (execute-enter ctx)
  "Run all the #:enter functions in the queue."
  (if (null? (context-queue ctx))
      ;; Prepare to leave
      (set-context-queue! ctx (context-stack ctx))
      (let ((t (car (context-queue ctx))))
        ;; Run the enter handler for the interceptor
        (try-enter ctx t)
        ;; Remove the current interceptor from the queue and add it to the stack
        (set-context-stack! ctx (cons t (context-stack ctx)))
        (set-context-queue! ctx (cdr (context-queue ctx)))
        (if (context-error ctx)
            ;; If an error was caught, abort the enter phase and set up to run the leave phase
            (begin
              (set-context-queue! ctx (context-stack ctx))
              (set-context-stack! ctx '()))
            ;; Otherwise, check for early termination or carry on down the chain
            (begin
              (check-terminators ctx)
              (execute-enter ctx))))))

(define (terminate-when ctx pred)
  "Add a predicate for a termination condition to exit the #:enter chain early."
  (set-context-terminators! ctx (cons pred (context-terminators ctx))))

(define (execute ctx interceptors)
  "Execute all the interceptors on the given context."
  (log-msg 'DEBUG "Enqueuing interceptors: " (map interceptor-name interceptors))
  (enqueue ctx interceptors)
  (log-msg 'DEBUG "Starting #:enter chain: " (map interceptor-name (context-queue ctx)))
  (execute-enter ctx)
  (log-msg 'DEBUG "Starting #:leave chain: " (map interceptor-name (context-queue ctx)))
  (execute-leave ctx)
  (and=> (context-error ctx) raise-exception))
